import json
import logging

from celery import bootsteps, Task, group


from collections import defaultdict
from lxml import etree

from inventory_provider.tasks.app import app
from inventory_provider.tasks.common import get_redis
from inventory_provider import config
from inventory_provider import constants
from inventory_provider import environment
from inventory_provider.db import db, opsdb, alarmsdb
from inventory_provider import snmp
from inventory_provider import juniper

# TODO: error callback (cf. http://docs.celeryproject.org/en/latest/userguide/calling.html#linking-callbacks-errbacks)  # noqa: E501

environment.setup_logging()


class InventoryTask(Task):

    config = None
    logger = None

    def __init__(self):
        pass

    # @staticmethod
    # def save_key(hostname, key, value):
    #     assert isinstance(value, str), \
    #         "sanity failure: expected string data as value"
    #     r = get_redis(InventoryTask.config)
    #     r.hset(
    #         name=hostname,
    #         key=key,
    #         value=value)
    #     InventoryTask.logger.debug(
    #         "saved %s, key %s" % (hostname, key))
    #     return "OK"

    @staticmethod
    def save_value(key, value):
        assert isinstance(value, str), \
            "sanity failure: expected string data as value"
        r = get_redis(InventoryTask.config)
        r.set(name=key, value=value)
        InventoryTask.logger.debug("saved %s" % key)
        return "OK"

    @staticmethod
    def save_value_json(key, data_obj):
        InventoryTask.save_value(
            key,
            json.dumps(data_obj))

    @staticmethod
    def save_value_etree(key, xml_doc):
        InventoryTask.save_value(
            key,
            etree.tostring(xml_doc, encoding='unicode'))

    # @staticmethod
    # def save_key_json(hostname, key, data_obj):
    #     InventoryTask.save_key(
    #         hostname,
    #         key,
    #         json.dumps(data_obj))
    #
    # @staticmethod
    # def save_key_etree(hostname, key, xml_doc):
    #     InventoryTask.save_key(
    #         hostname,
    #         key,
    #         etree.tostring(xml_doc, encoding='unicode'))


class WorkerArgs(bootsteps.Step):
    def __init__(self, worker, config_filename, **options):
        with open(config_filename) as f:
            InventoryTask.config = config.load(f)
        InventoryTask.logger = logging.getLogger(constants.TASK_LOGGER_NAME)


# interfaces_key = "interface_services"
# equipment_locations_key = "equipment_locations"
# service_child_to_parents_key = "child_to_parent_circuit_relations"
# service_parent_to_children_key = "parent_to_children_circuit_relations"
# interface_status_key = "interface_statuses"


def worker_args(parser):
    parser.add_argument(
        "--config_filename",
        dest="config_filename",
        action='store',
        type=str,
        help="Configuration filename")


app.user_options['worker'].add(worker_args)
app.steps['worker'].add(WorkerArgs)


@app.task(bind=InventoryTask)
def snmp_refresh_interfaces(self, hostname, community):
    logger = logging.getLogger(constants.TASK_LOGGER_NAME)
    logger.debug('STARTING: snmp_refresh_interfaces(%r, %r)'
                 % (hostname, community))

    InventoryTask.save_value_json(
        'snmp-interfaces:' + hostname,
        list(snmp.get_router_interfaces(
            hostname,
            community,
            InventoryTask.config)))

    logger.debug('FINISHED: snmp_refresh_interfaces(%r, %r)'
                 % (hostname, community))


@app.task(bind=InventoryTask)
def netconf_refresh_config(self, hostname):
    logger = logging.getLogger(constants.TASK_LOGGER_NAME)
    logger.debug('STARTING: netconf_refresh_config(%r)' % hostname)

    InventoryTask.save_value_etree(
        'netconf:' + hostname,
        juniper.load_config(hostname, InventoryTask.config["ssh"]))

    logger.debug('FINISHED: netconf_refresh_config(%r)' % hostname)


# @app.task(bind=InventoryTask)
# def update_alarmsdb_cache(self):
#     logger = logging.getLogger(constants.TASK_LOGGER_NAME)
#     logger.debug('STARTING: update_alarmsdb_cache')
#
#     with db.connection(InventoryTask.config["alarms-db"]) as cx:
#         for table_name, data in alarmsdb.load_cache(cx):
#             InventoryTask.save_value_json('alarmsdb:%s' % table_name, data)
#
#     logger.debug('FINISHED: update_alarmsdb_cache')


@app.task()
def update_interfaces_to_services():
    interface_services = defaultdict(list)
    with db.connection(InventoryTask.config["ops-db"]) as cx:
        for service in opsdb.get_circuits(cx):
            equipment_interface = '%s:%s' % (
                service['equipment'], service['interface_name'])
            interface_services[equipment_interface].append(service)

    r = get_redis(InventoryTask.config)
    for key in r.scan_iter('opsdb:interface_services:*'):
        r.delete(key)
    for equipment_interface, services in interface_services.items():
        r.set(
            'opsdb:interface_services:' + equipment_interface,
            json.dumps(services))


@app.task()
def update_equipment_locations():
    r = get_redis(InventoryTask.config)
    for key in r.scan_iter('opsdb:location:*'):
        r.delete(key)
    with db.connection(InventoryTask.config["ops-db"]) as cx:
        for ld in opsdb.get_equipment_location_data(cx):
            r.set('opsdb:location:%s' % ld['equipment_name'], json.dumps(ld))


@app.task()
def update_circuit_hierarchy():
    # TODO: integers are not JSON keys
    with db.connection(InventoryTask.config["ops-db"]) as cx:
        child_to_parents = defaultdict(list)
        parent_to_children = defaultdict(list)
        for relation in opsdb.get_circuit_hierarchy(cx):
            parent_id = relation["parent_circuit_id"]
            child_id = relation["child_circuit_id"]
            parent_to_children[parent_id].append(relation)
            child_to_parents[child_id].append(relation)

        r = get_redis(InventoryTask.config)
        for key in r.scan_iter('opsdb:services:parents:*'):
            r.delete(key)
        for cid, parents in child_to_parents.items():
            r.set('opsdb:services:parents:%d' % cid, json.dumps(parents))

        for key in r.scan_iter('opsdb:services:children:*'):
            r.delete(key)
        for cid, children in child_to_parents.items():
            r.set('opsdb:services:children:%d' % cid, json.dumps(children))


@app.task()
def update_interface_statuses():
    with db.connection(InventoryTask.config["ops-db"]) as cx:
        services = opsdb.get_circuits(cx)
    with db.connection(InventoryTask.config["alarms-db"]) as cx:
        with db.cursor(cx) as csr:
            for service in services:
                key = 'alarmsdb:interface_status:%s:%s' \
                      % (service['equipment'], service['interface_name'])
                status = alarmsdb.get_last_known_interface_status(
                    csr,
                    service["equipment"],
                    service["interface_name"])
                InventoryTask.save_value(key, status)


@app.task()
def update_inventory_system_cache():
    logger = logging.getLogger(constants.TASK_LOGGER_NAME)
    logger.error('HERE: update_inventory_system_cache')

    g = group([
        update_interfaces_to_services.s(),
        update_circuit_hierarchy.s(),
        update_equipment_locations.s(),
        update_interface_statuses.s()
    ])
    g.apply_async()
